#!/usr/bin/env python

from nidibot.server_provider.nitrado_server_provider import NitradoServerProvider
from nidibot.server_provider.server_provider_base import ServerProviderConfiguration


class ServerProviderFactory:
    __supported_server_providers = {
        "nitrado": NitradoServerProvider,
    }

    @staticmethod
    def create(
        configuration: ServerProviderConfiguration,
        root_backup_path: str,
        notify_callback,
    ):
        if configuration.type not in ServerProviderFactory.__supported_server_providers:
            if not configuration.type:
                raise ValueError("Empty bot type provided!")

            raise ValueError("Unknown bot type provided!")

        return ServerProviderFactory.__supported_server_providers[configuration.type](
            configuration=configuration,
            backup_directory=root_backup_path,
            notify_callback=notify_callback,
        )

    @staticmethod
    def create_all(
        configuration_list: list, root_backup_path: str, notify_callback
    ) -> list:
        server_providers: list = []

        for configuration in configuration_list:
            server_providers.append(
                ServerProviderFactory.create(
                    configuration=configuration,
                    root_backup_path=root_backup_path,
                    notify_callback=notify_callback,
                )
            )

        return server_providers



#!/usr/bin/env python

import os
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field


@dataclass
class ServerProviderNotificationConfiguration:
    on_new_server: bool = False
    on_status_change: bool = False
    on_address_change: bool = False
    on_version_change: bool = False
    on_update_available_change: bool = False


@dataclass
class ServerProviderConfiguration:
    type: str = ""
    token: str = ""
    timeout_seconds: int = 10
    polling_seconds: int = 5
    notifications: ServerProviderNotificationConfiguration = field(
        default_factory=ServerProviderNotificationConfiguration
    )


@dataclass
class ServerStatus:
    game_name: str = ""
    version: str = ""
    address: str = ""
    status: str = ""
    available_until: str = ""
    update_available: bool = False
    players_limit: int = 0
    players_connected: int = 0
    player_names: list = field(default_factory=list)


class ServerProviderBase(ABC):
    def __init__(
        self,
        configuration: ServerProviderConfiguration,
        backup_directory: str,
        notify_callback,
    ):
        if not configuration.token:
            raise ValueError("API token value is required for nidibot start!")

        if not backup_directory:
            raise ValueError("Backup directory value is required for nidibot start!")

        self._configuration = configuration
        self._root_backup_directory = backup_directory
        self._notify_callback = notify_callback

    @abstractmethod
    def _poll(self) -> None:
        pass

    def _get_backup_directory_path(self, game_name: str, server_id: str) -> str:
        return os.path.join(
            self._root_backup_directory, self.name().lower(), game_name, server_id
        )

    def _wait_for_status(
        self,
        server_id: str,
        required_status: str,
        timeout_seconds: int = 60,
        wait_step_seconds: int = 1,
    ) -> bool:
        end_time = time.time() + timeout_seconds
        while time.time() < end_time:
            status = self.status(server_id)
            if status.status == required_status:
                return True

            time.sleep(wait_step_seconds)

        return False

    @abstractmethod
    def get_servers(self) -> list:
        pass

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def status(self, server_id: str = "") -> ServerStatus:
        pass

    @abstractmethod
    def start(self, server_id: str = "") -> bool:
        pass

    @abstractmethod
    def stop(self, server_id: str = "") -> bool:
        pass

    @abstractmethod
    def restart(self, server_id: str = "") -> bool:
        pass

    @abstractmethod
    def create_backup(self, server_id: str = "") -> bool:
        pass

    @abstractmethod
    def restore_backup(self, server_id: str = "", filepath: str = "") -> bool:
        pass

    @abstractmethod
    def list_backups(self, server_id: str = "") -> list:
        pass



#!/usr/bin/env python

import glob
import json
import logging
import os
import pathlib
import shutil
import subprocess
import tempfile
import threading
import time
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

import ftputil  # type: ignore
import requests  # type: ignore

from nidibot.bots.bot_base import BackupDescription
from nidibot.server_provider.game_server import GameServer
from nidibot.server_provider.server_provider_base import (
    ServerProviderBase,
    ServerProviderConfiguration,
    ServerStatus,
)


@dataclass
class NitradoService:
    id: str = ""
    short_name: str = ""
    available_until: str = ""


@dataclass
class NitradoFtpConfiguration:
    hostname: str = ""
    port: int = 21
    username: str = ""
    password: str = ""


@dataclass
class NitradoMysqlConfiguration:
    hostname: str = ""
    port: int = 21
    username: str = ""
    password: str = ""
    database: str = ""


@dataclass
class NitradoServerInformation:
    id: str = ""
    short_name: str = ""
    status: ServerStatus = field(default_factory=ServerStatus)
    ftp: NitradoFtpConfiguration = field(default_factory=NitradoFtpConfiguration)
    mysql: NitradoMysqlConfiguration = field(default_factory=NitradoMysqlConfiguration)


class NitradoServerProvider(ServerProviderBase):
    def __init__(
        self,
        configuration: ServerProviderConfiguration,
        backup_directory: str,
        notify_callback,
    ):
        super().__init__(
            configuration=configuration,
            backup_directory=backup_directory,
            notify_callback=notify_callback,
        )
        self.__servers: Dict[str, NitradoServerInformation] = {}

        logging.basicConfig(
            level=logging.DEBUG,
            format="%(asctime)s:%(levelname)s:%(funcName)s %(message)s",
            handlers=[logging.StreamHandler()],
        )

        self.__verify_api_version()

        self.__data_received_event = threading.Event()
        self.__stopping_event = threading.Event()

        self.__polling_thread = threading.Thread(target=self._poll)
        self.__polling_thread.daemon = True
        self.__polling_thread.start()

        self.__data_received_event.wait()

    def __verify_api_version(self) -> None:
        response = requests.get(
            "https://api.nitrado.net/version",
            timeout=self._configuration.timeout_seconds,
        )

        logging.debug(
            "Response received, status code: %d, content: %s.",
            response.status_code,
            response.content.decode("utf-8"),
        )

        content_json = json.loads(response.content.decode("utf-8"))

        if response.status_code != 200:
            raise ValueError("Failed reading Nitrado API version!")

        if content_json["status"] != "success":
            raise ValueError("Failed reading Nitrado API version!")

        # Warning: Nitrado always changes last part of provided API version.
        expected_api_version = "nitrapi-1471"
        if expected_api_version not in content_json["message"]:
            logging.warning(
                "Nitrado API version was changed! Expected: %s, actual: %s.",
                expected_api_version,
                content_json["message"],
            )

    def __download_ftp_folder(
        self, ftp_server, local_path: str, remote_path: str, ignore_folders: list
    ):
        for root_path, folders, files in ftp_server.walk(top=remote_path, topdown=True):
            root_folder_name = os.path.basename(root_path)
            if root_folder_name in ignore_folders:
                return

            for filename in files:
                remote_filepath = ftp_server.path.join(root_path, filename)
                local_filepath = os.path.join(local_path, remote_filepath)
                local_parent_folder = pathlib.Path(local_filepath).parent.absolute()
                pathlib.Path(local_parent_folder).mkdir(parents=True, exist_ok=True)
                ftp_server.download(remote_filepath, local_filepath)
                logging.debug(
                    "Downloaded '%s' to '%s'.", remote_filepath, local_filepath
                )

            for folder_name in folders:
                self.__download_ftp_folder(
                    ftp_server=ftp_server,
                    local_path=local_path,
                    remote_path=os.path.join(root_path, folder_name),
                    ignore_folders=ignore_folders,
                )

            return

    def __upload_ftp_folder(
        self, ftp_server, local_path: str, remote_path: str, ignore_folders: list
    ):
        for root_path, folders, files in os.walk(top=local_path, topdown=True):
            root_folder_name = os.path.basename(root_path)
            if root_folder_name in ignore_folders:
                return

            for filename in files:
                local_filepath = os.path.join(root_path, filename)
                remote_filepath = ftp_server.path.join(remote_path, filename)

                if remote_path != ".":
                    ftp_server.makedirs(remote_path, exist_ok=True)

                ftp_server.upload(local_filepath, remote_filepath)
                logging.debug("Uploaded '%s' to '%s'.", local_filepath, remote_filepath)

            for folder_name in folders:
                self.__upload_ftp_folder(
                    ftp_server=ftp_server,
                    local_path=os.path.join(root_path, folder_name),
                    remote_path=os.path.join(remote_path, folder_name),
                    ignore_folders=ignore_folders,
                )

            return

    def __get_services(self) -> list:
        service_list = []

        try:
            headers = {"Authorization": self._configuration.token}
            response = requests.get(
                "https://api.nitrado.net/services",
                timeout=self._configuration.timeout_seconds,
                headers=headers,
            )

            logging.debug(
                "Response received, status code: %d.",
                response.status_code,
            )

            content_json = json.loads(response.content.decode("utf-8"))
            data_service_list = content_json["data"]["services"]
            for data_service in data_service_list:
                service: NitradoService = NitradoService()

                service.id = data_service["id"]
                service.short_name = data_service["details"]["folder_short"]
                service.available_until = str(data_service["suspend_date"]).replace(
                    "T", " "
                )

                service_list.append(service)

        except Exception as exception:
            logging.exception(exception)

        return service_list

    def _poll(self) -> None:
        logging.debug("Polling thread started.")

        while not self.__stopping_event.is_set():
            servers: Dict[str, NitradoServerInformation] = {}
            services = self.__get_services()
            for service in services:
                try:
                    headers = {"Authorization": self._configuration.token}
                    response = requests.get(
                        f"https://api.nitrado.net/services/{service.id}/gameservers",
                        timeout=self._configuration.timeout_seconds,
                        headers=headers,
                    )

                    logging.debug(
                        "Response received, status code: %d.",
                        response.status_code,
                    )

                    content_json = json.loads(response.content.decode("utf-8"))
                    gameserver_dict = content_json["data"]["gameserver"]
                    query_dict = gameserver_dict["query"]

                    server: NitradoServerInformation = NitradoServerInformation()
                    server.id = str(service.id)
                    server.short_name = service.short_name
                    server.status.game_name = gameserver_dict["game_human"]
                    server.status.update_available = (
                        gameserver_dict["game_specific"]["update_status"]
                        != "up_to_date"
                    )
                    server.status.available_until = service.available_until

                    if gameserver_dict["status"] == "started":
                        server.status.status = "online"
                    elif gameserver_dict["status"] == "stopped":
                        server.status.status = "offline"
                    else:
                        server.status.status = str(gameserver_dict["status"])

                    if len(query_dict) > 0:
                        server.status.version = gameserver_dict["query"]["version"]
                        server.status.address = gameserver_dict["query"]["connect_ip"]
                        server.status.players_limit = int(
                            gameserver_dict["query"]["player_max"]
                        )
                        server.status.players_connected = int(
                            gameserver_dict["query"]["player_current"]
                        )
                        server.status.player_names = gameserver_dict["query"]["players"]
                    else:
                        server.status.address = (
                            str(gameserver_dict["ip"])
                            + ":"
                            + str(gameserver_dict["query_port"])
                        )
                        server.status.players_limit = int(gameserver_dict["slots"])
                        server.status.player_names = []

                    ftp_dict = content_json["data"]["gameserver"]["credentials"]["ftp"]

                    server.ftp.hostname = ftp_dict["hostname"]
                    server.ftp.port = int(ftp_dict["port"])
                    server.ftp.username = ftp_dict["username"]
                    server.ftp.password = ftp_dict["password"]

                    mysql_dict = content_json["data"]["gameserver"]["credentials"][
                        "mysql"
                    ]

                    server.mysql.hostname = mysql_dict["hostname"]
                    server.mysql.port = int(mysql_dict["port"])
                    server.mysql.username = mysql_dict["username"]
                    server.mysql.password = mysql_dict["password"]
                    server.mysql.database = mysql_dict["database"]

                    servers[server.id] = server

                except Exception as exception:
                    logging.exception(exception)

            #
            # Check for changes in status and notify.
            #
            if len(self.__servers) > 0:
                for server_id, server in servers.items():
                    title = f"{server.status.game_name}"
                    if server.status.version:
                        title += f" ({server.status.version})"

                    title += f" - {server.status.address}"

                    if (
                        self._configuration.notifications.on_new_server
                        and server_id not in self.__servers
                    ):
                        self._notify_callback(
                            title, "New game server appeared, please configure it."
                        )

                    if (
                        self._configuration.notifications.on_status_change
                        and server.status.status
                        != self.__servers[server_id].status.status
                    ):
                        self._notify_callback(
                            title,
                            f"Status changed from '{self.__servers[server_id].status.status}' to "
                            f"'{server.status.status}'.",
                        )

                    if (
                        self._configuration.notifications.on_address_change
                        and server.status.address
                        != self.__servers[server_id].status.address
                    ):
                        self._notify_callback(
                            title,
                            f"Address from '{self.__servers[server_id].status.address}' to "
                            f"'{server.status.address}'.",
                        )

                    if (
                        self._configuration.notifications.on_version_change
                        and server.status.version
                        != self.__servers[server_id].status.version
                    ):
                        self._notify_callback(
                            title,
                            f"Version from '{self.__servers[server_id].status.version}' to "
                            f"'{server.status.version}'.",
                        )

                    if (
                        self._configuration.notifications.on_update_available_change
                        and server.status.update_available
                        != self.__servers[server_id].status.update_available
                    ):
                        self._notify_callback(
                            title,
                            "Update is available, please restart server.",
                        )

            self.__servers = servers

            self.__data_received_event.set()
            self.__stopping_event.wait(self._configuration.polling_seconds)

    def name(self) -> str:
        return "Nitrado"

    def get_servers(self) -> List[GameServer]:
        servers: List[GameServer] = []

        for key, value in self.__servers.items():
            name = f"{value.short_name}-{value.status.address}"
            servers.append(
                GameServer(server_provider=self, server_id=key, server_name=name)
            )

        return servers

    def status(self, server_id: str = "") -> ServerStatus:
        return self.__servers[server_id].status

    def start(self, server_id: str = "") -> bool:
        return self.restart(server_id)

    def stop(self, server_id: str = "") -> bool:
        try:
            headers = {"Authorization": self._configuration.token}
            response = requests.post(
                f"https://api.nitrado.net/services/{server_id}/gameservers/stop",
                timeout=self._configuration.timeout_seconds,
                headers=headers,
            )

            logging.debug(
                "Response received, status code: %d, content: %s.",
                response.status_code,
                response.content.decode("utf-8"),
            )

            content_json = json.loads(response.content.decode("utf-8"))

            return response.status_code == 200 and content_json["status"] == "success"

        except Exception as exception:
            logging.exception(exception)
            return False

    def restart(self, server_id: str = "") -> bool:
        try:
            headers = {"Authorization": self._configuration.token}
            response = requests.post(
                f"https://api.nitrado.net/services/{server_id}/gameservers/restart",
                timeout=self._configuration.timeout_seconds,
                headers=headers,
            )

            logging.debug(
                "Response received, status code: %d, content: %s.",
                response.status_code,
                response.content.decode("utf-8"),
            )

            content_json = json.loads(response.content.decode("utf-8"))

            return response.status_code == 200 and content_json["status"] == "success"

        except Exception as exception:
            logging.exception(exception)
            return False

    def create_backup(self, server_id: str = "") -> bool:
        try:
            start_time = time.time()
            with tempfile.TemporaryDirectory() as temp_folder_path:
                datetime_str = datetime.now().strftime("%Y%m%d_%H%M%S")
                local_path = os.path.join(temp_folder_path, datetime_str)
                pathlib.Path(local_path).mkdir(parents=True, exist_ok=True)

                #
                # Copy game server files via FTP to temporary folder.
                #
                files_path = os.path.join(local_path, "files")
                with ftputil.FTPHost(
                    self.__servers[server_id].ftp.hostname,
                    self.__servers[server_id].ftp.username,
                    self.__servers[server_id].ftp.password,
                ) as ftp:
                    self.__download_ftp_folder(
                        ftp_server=ftp,
                        local_path=files_path,
                        remote_path=".",
                        ignore_folders=["Crashes", "CrashReportClient"],
                    )

                #
                # Save MySQL database to temporary folder.
                #
                mysqldump_path = shutil.which("mysqldump")
                if mysqldump_path is not None:
                    database_name = self.__servers[server_id].mysql.database
                    mysql_folder_path = os.path.join(local_path, "mysql")
                    pathlib.Path(mysql_folder_path).mkdir(parents=True, exist_ok=True)
                    database_filepath = os.path.join(
                        mysql_folder_path, database_name + ".sql"
                    )

                    mysqldump_command = (
                        f"{mysqldump_path} --host {self.__servers[server_id].mysql.hostname} "
                        f"--port {str(self.__servers[server_id].mysql.port)} "
                        f"--user {self.__servers[server_id].mysql.username} "
                        f"-p{self.__servers[server_id].mysql.password} "
                        f"{database_name} > {database_filepath}"
                    )

                    # logging.debug("Executing: '%s'.", mysql_command)

                    with subprocess.Popen(
                        mysqldump_command,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        shell=True,
                        encoding="utf-8",
                        universal_newlines=True,
                        cwd=local_path,
                    ) as process:
                        for stdout_line in iter(process.stdout.readline, ""):  # type: ignore
                            logging.info(stdout_line.replace("\n", ""))
                        process.stdout.close()  # type: ignore
                        _ = process.wait()

                else:
                    logging.warning(
                        "No mysqldump is available in the system, please install it."
                    )

                #
                # Create ZIP archive.
                #
                backup_directory = self._get_backup_directory_path(
                    game_name=self.__servers[server_id].short_name, server_id=server_id
                )
                pathlib.Path(backup_directory).mkdir(parents=True, exist_ok=True)

                shutil.make_archive(
                    os.path.join(backup_directory, datetime_str), "zip", local_path
                )

                logging.debug(
                    "Created backup archive at path: '%s'.",
                    os.path.join(backup_directory, datetime_str),
                )

            end_time = time.time()
            logging.debug("Backup creation took %s.", end_time - start_time)

        except Exception as exception:
            logging.exception(exception)
            return False

        return True

    def restore_backup(self, server_id: str = "", filepath: str = "") -> bool:
        status = self.status(server_id=server_id)
        if status.status != "offline":
            if not self.stop(server_id=server_id):
                return False

        if not self._wait_for_status(
            server_id=server_id, required_status="offline", timeout_seconds=60
        ):
            return False

        start_time = time.time()
        with tempfile.TemporaryDirectory() as temp_folder_path:
            shutil.unpack_archive(filepath, temp_folder_path)

            #
            # Copy game server files via FTP to temporary folder.
            #
            files_path = os.path.join(temp_folder_path, "files")
            with ftputil.FTPHost(
                self.__servers[server_id].ftp.hostname,
                self.__servers[server_id].ftp.username,
                self.__servers[server_id].ftp.password,
            ) as ftp:
                self.__upload_ftp_folder(
                    ftp_server=ftp,
                    local_path=files_path,
                    remote_path=".",
                    ignore_folders=["Crashes", "CrashReportClient"],
                )

            #
            # Save MySQL database to temporary folder.
            #
            mysql_path = shutil.which("mysql")
            if mysql_path is not None:
                database_name = self.__servers[server_id].mysql.database
                database_filepath = os.path.join(
                    temp_folder_path, "mysql", database_name + ".sql"
                )

                if os.path.exists(database_filepath):
                    mysql_command = (
                        f"{mysql_path} --host={self.__servers[server_id].mysql.hostname} "
                        f"--port={str(self.__servers[server_id].mysql.port)} "
                        f"--user={self.__servers[server_id].mysql.username} "
                        f"--password={self.__servers[server_id].mysql.password} "
                        f"{database_name} < {database_filepath}"
                    )

                    # logging.debug("Executing: '%s'.", mysql_command)

                    with subprocess.Popen(
                        args=mysql_command,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        shell=True,
                        encoding="utf-8",
                        universal_newlines=True,
                        cwd=temp_folder_path,
                    ) as process:
                        for stdout_line in iter(process.stdout.readline, ""):  # type: ignore
                            logging.info(stdout_line.replace("\n", ""))
                        process.stdout.close()  # type: ignore
                        _ = process.wait()
                else:
                    logging.warning("No MySQL database file found, nothing to restore!")

            else:
                logging.warning(
                    "No mysql is available in the system, please install it."
                )

        # Start server back if it was online before restore.
        if status.status != "offline":
            self.start(server_id=server_id)

        end_time = time.time()
        logging.debug("Backup restoring took %s.", end_time - start_time)

        return True

    def list_backups(self, server_id: str = "") -> list:
        backup_directory = self._get_backup_directory_path(
            game_name=self.__servers[server_id].short_name, server_id=server_id
        )

        wildcard_path = os.path.join(backup_directory, "*")
        file_list = glob.glob(wildcard_path)

        backups: List[BackupDescription] = []
        for filepath in file_list:
            pathlib_filepath = pathlib.Path(filepath)
            extension = pathlib_filepath.suffix
            filename = pathlib_filepath.name.replace(extension, "")
            filename_parts = filename.split("_")
            if len(filename_parts) < 2:
                continue

            date_str = filename_parts[-2]
            time_str = filename_parts[-1]

            date_val = datetime.strptime(date_str, "%Y%m%d")
            time_val = datetime.strptime(time_str, "%H%M%S")

            backup_description: BackupDescription = BackupDescription()
            backup_description.filepath = filepath
            backup_description.readable_name = (
                f"{date_val.strftime('%Y-%m-%d')} {time_val.strftime('%H:%M:%S')}"
            )

            backups.append(backup_description)

        backups.sort()
        backups.reverse()

        return backups



#!/usr/bin/env python

import logging
import typing
from datetime import date, datetime
from typing import List

import hikari
import hikari.colors
import hikari.errors
import lightbulb
from hikari.api import MessageActionRowBuilder
from lightbulb.ext import tasks

from nidibot.bots.bot_base import (
    BackupDescription,
    BotBase,
    BotConfiguration,
    BotForwardMessage,
)
from nidibot.server_provider.game_server import GameServer


class DiscordBot(BotBase):
    def __init__(self, configuration: BotConfiguration, game_servers: List[GameServer]):
        super().__init__(configuration=configuration, game_servers=game_servers)

        self.__bot = lightbulb.BotApp(token=self._configuration.token)
        tasks.load(self.__bot)

        self.__color_green = 0x37CB78
        self.__color_orange = 0xE67E22
        self.__color_red = 0xE64A42

        @self.__bot.listen(hikari.StartedEvent)
        async def on_started(_) -> None:
            logging.info("Discord bot started and connected.")

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self._game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="status",
            description="Provides extended information about game server status.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def status(ctx) -> None:
            if (
                len(self._configuration.allowed_channels) > 0
                and str(ctx.channel_id) not in self._configuration.allowed_channels
            ):
                logging.error(
                    "Called 'status' by '%s' in not allowed channel '%s'.",
                    ctx.author,
                    ctx.channel_id,
                )
                return

            logging.debug("Called 'status' by '%s'.", ctx.author)

            game_server = self._get_game_server(ctx.options.name)
            server_status = game_server.status()

            title = self._get_response_title(game_server)

            if server_status.status == "online":
                status_emoji = self._emoji_ok
                status_color = hikari.colors.Color(self.__color_green)
            elif server_status.status == "offline":
                status_emoji = self._emoji_bad
                status_color = hikari.colors.Color(self.__color_red)
            elif server_status.status == "restarting":
                status_emoji = self._emoji_attention
                status_color = hikari.colors.Color(self.__color_orange)
            else:
                status_emoji = self._emoji_unknown
                status_color = hikari.colors.Color(self.__color_orange)

            embed = hikari.Embed(
                title=title,
                color=status_color,
            )

            embed.add_field(
                name="Address:", value=f"`{server_status.address}`", inline=True
            )
            embed.add_field(
                name="Status:",
                value=f"{status_emoji} {server_status.status}",
                inline=True,
            )

            players = (
                f"{server_status.players_connected} / {server_status.players_limit}"
            )
            if len(server_status.player_names) > 0:
                players += f" ({server_status.player_names})"
            embed.add_field(name="Players:", value=f"{players}", inline=True)

            update_emoji = ""
            if server_status.update_available:
                update_emoji = self._emoji_attention
            else:
                update_emoji = self._emoji_ok

            date_time = server_status.available_until.split(" ")
            date_parts = date_time[0].split("-")
            d0 = date(datetime.now().year, datetime.now().month, datetime.now().day)
            d1 = date(int(date_parts[0]), int(date_parts[1]), int(date_parts[2]))
            delta = d1 - d0

            embed.add_field(
                name="Available until:",
                value=f"{server_status.available_until} ({delta.days} days left)",
                inline=True,
            )
            embed.add_field(
                name="Update available:",
                value=f"{update_emoji} {'yes' if server_status.update_available else 'no'}",
                inline=True,
            )

            await ctx.respond(embed=embed)

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self._game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="start",
            description="Starts server if it is offline, restarts server if it is online.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def start(ctx) -> None:
            if (
                len(self._configuration.allowed_channels) > 0
                and str(ctx.channel_id) not in self._configuration.allowed_channels
            ):
                logging.error(
                    "Called 'start' by '%s' in not allowed channel '%s'.",
                    ctx.author,
                    ctx.channel_id,
                )
                return

            logging.debug("Called 'start' by '%s'.", ctx.author)

            game_server = self._get_game_server(ctx.options.name)
            title = self._get_response_title(game_server)

            user = str(ctx.author)
            if user not in self._configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description=f"Sorry but you don't have rights to call this command! {self._emoji_no_access}",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=f"{self._emoji_attention} Starting server!",
                color=hikari.colors.Color(self.__color_red),
            )
            await ctx.respond(embed=embed)

            game_server.start()

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self._game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="stop",
            description="Stops server if it is online.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def stop(ctx) -> None:
            if (
                len(self._configuration.allowed_channels) > 0
                and str(ctx.channel_id) not in self._configuration.allowed_channels
            ):
                logging.error(
                    "Called 'stop' by '%s' in not allowed channel '%s'.",
                    ctx.author,
                    ctx.channel_id,
                )
                return

            logging.debug("Called 'stop' by '%s'.", ctx.author)

            game_server = self._get_game_server(ctx.options.name)
            title = self._get_response_title(game_server)

            user = str(ctx.author)
            if user not in self._configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description=f"Sorry but you don't have rights to call this command! {self._emoji_no_access}",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=f"{self._emoji_attention} Stopping server!",
                color=hikari.colors.Color(self.__color_red),
            )
            await ctx.respond(embed=embed)

            game_server.stop()

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self._game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="restart",
            description="Restarts server if it is online, starts server if it is offline.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def restart(ctx) -> None:
            if (
                len(self._configuration.allowed_channels) > 0
                and str(ctx.channel_id) not in self._configuration.allowed_channels
            ):
                logging.error(
                    "Called 'restart' by '%s' in not allowed channel '%s'.",
                    ctx.author,
                    ctx.channel_id,
                )
                return

            logging.debug("Called 'restart' by '%s'.", ctx.author)

            game_server = self._get_game_server(ctx.options.name)
            title = self._get_response_title(game_server)

            user = str(ctx.author)
            if user not in self._configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description=f"Sorry but you don't have rights to call this command! {self._emoji_no_access}",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=f"{self._emoji_attention} Restarting server!",
                color=hikari.colors.Color(self.__color_red),
            )
            await ctx.respond(embed=embed)

            game_server.restart()

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self._game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="backup_create",
            description="Creates backup of game server files and uploads them to storage.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def backup_create(ctx) -> None:
            if (
                len(self._configuration.allowed_channels) > 0
                and str(ctx.channel_id) not in self._configuration.allowed_channels
            ):
                logging.error(
                    "Called 'backup_create' by '%s' in not allowed channel '%s'.",
                    ctx.author,
                    ctx.channel_id,
                )
                return

            logging.debug("Called 'backup_create' by '%s'.", ctx.author)

            game_server = self._get_game_server(ctx.options.name)
            title = self._get_response_title(game_server)

            user = str(ctx.author)
            if user not in self._configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description=f"Sorry but you don't have rights to call this command! {self._emoji_no_access}",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=f"{self._emoji_attention} Started creating backup of the server, please wait.",
                color=hikari.colors.Color(self.__color_orange),
            )
            await ctx.respond(embed=embed)

            if game_server.create_backup():
                embed = hikari.Embed(
                    title=title,
                    description=f"{self._emoji_ok} Backup was created successfully!",
                    color=hikari.colors.Color(self.__color_green),
                )
            else:
                embed = hikari.Embed(
                    title=title,
                    description=f"{self._emoji_bad} Backup creation failed, please check bot logs!",
                    color=hikari.colors.Color(self.__color_red),
                )

            await ctx.respond(embed=embed)

        async def create_backup_buttons(
            bot: lightbulb.BotApp, backups: List[BackupDescription]
        ) -> typing.Iterable[MessageActionRowBuilder]:

            rows: typing.List[MessageActionRowBuilder] = []
            row = bot.rest.build_message_action_row()

            buttons_added_to_row = 0
            for backup_description in backups:
                if buttons_added_to_row % 5 == 0 and buttons_added_to_row != 0:
                    rows.append(row)
                    row = bot.rest.build_message_action_row()
                    buttons_added_to_row = 0

                row.add_interactive_button(
                    hikari.ButtonStyle.SECONDARY,
                    backup_description.filepath,
                    label=backup_description.readable_name,
                )

                buttons_added_to_row += 1

            rows.append(row)

            return rows

        async def handle_backup_restore(
            ctx,
            message: hikari.Message,
            title: str,
            backups: List[BackupDescription],
            game_server: GameServer,
        ) -> None:

            with ctx.bot.stream(hikari.InteractionCreateEvent, 120).filter(
                lambda e: (
                    isinstance(e.interaction, hikari.ComponentInteraction)
                    and e.interaction.user == ctx.author
                    and e.interaction.message == message
                )
            ) as stream:
                async for event in stream:
                    interaction: hikari.ComponentInteraction = event.interaction  # type: ignore
                    filename = interaction.custom_id
                    backup_description = next(
                        (x for x in backups if x.filepath == filename), None
                    )
                    if backup_description is None:
                        return

                    embed = hikari.Embed(
                        title=title,
                        color=hikari.colors.Color(self.__color_orange),
                        description=f"Selected backup: {backup_description.readable_name}",
                    )

                    try:
                        await interaction.create_initial_response(
                            hikari.ResponseType.MESSAGE_UPDATE,
                            embed=embed,
                            components=[],
                        )
                    except hikari.NotFoundError:
                        await interaction.edit_initial_response(
                            embed=embed, components=[]
                        )

                    embed = hikari.Embed(
                        title=title,
                        description=f"{self._emoji_attention} Started restoring backup from {backup_description.readable_name}, please wait.",
                        color=hikari.colors.Color(self.__color_orange),
                    )
                    await ctx.respond(embed=embed)

                    if game_server.restore_backup(backup_description.filepath):
                        embed = hikari.Embed(
                            title=title,
                            description=f"{self._emoji_ok} Backup from {backup_description.readable_name} was restored successfully!",
                            color=hikari.colors.Color(self.__color_green),
                        )
                    else:
                        embed = hikari.Embed(
                            title=title,
                            description=f"{self._emoji_bad} Restoring backup from {backup_description.readable_name} failed, please check bot logs!",
                            color=hikari.colors.Color(self.__color_red),
                        )

                    await ctx.respond(embed=embed)

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self._game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="backup_restore",
            description="Restores specific backup on a game server.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def backup_restore(ctx) -> None:
            if (
                len(self._configuration.allowed_channels) > 0
                and str(ctx.channel_id) not in self._configuration.allowed_channels
            ):
                logging.error(
                    "Called 'backup_restore' by '%s' in not allowed channel '%s'.",
                    ctx.author,
                    ctx.channel_id,
                )
                return

            logging.debug("Called 'backup_restore' by '%s'.", ctx.author)

            game_server = self._get_game_server(ctx.options.name)
            title = self._get_response_title(game_server)

            user = str(ctx.author)
            if user not in self._configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description=f"Sorry but you don't have rights to call this command! {self._emoji_no_access}",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            backups = game_server.list_backups()
            if len(backups) > 0:
                self._backups[ctx.options.name] = backups

                embed = hikari.Embed(
                    title=title,
                    description="Select a backup:",
                    color=hikari.colors.Color(self.__color_orange),
                )

                backup_buttons = await create_backup_buttons(ctx.bot, backups)
                response = await ctx.respond(
                    embed=embed,
                    components=backup_buttons,
                )

                message = await response.message()
                await handle_backup_restore(ctx, message, title, backups, game_server)

            else:
                logging.warning("No backups available!")
                embed = hikari.Embed(
                    title=title,
                    description=f"{self._emoji_bad} No backups available!",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self._game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="backup_list",
            description="Lists available backups of specific game server.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def backup_list(ctx) -> None:
            if (
                len(self._configuration.allowed_channels) > 0
                and str(ctx.channel_id) not in self._configuration.allowed_channels
            ):
                logging.error(
                    "Called 'backup_list' by '%s' in not allowed channel '%s'.",
                    ctx.author,
                    ctx.channel_id,
                )
                return

            logging.debug("Called 'backup_list' by '%s'.", ctx.author)

            game_server = self._get_game_server(ctx.options.name)
            title = self._get_response_title(game_server)

            self._backups[ctx.options.name] = game_server.list_backups()

            backup_sum_message = "**Available backups:**\n"
            for backup in self._backups[ctx.options.name]:
                backup_sum_message += f"* {backup.readable_name}\n"

            embed = hikari.Embed(
                title=title,
                description=backup_sum_message,
                color=hikari.colors.Color(self.__color_orange),
            )
            await ctx.respond(embed=embed)

        @tasks.task(s=self._configuration.notify_polling_seconds, auto_start=True)
        async def notify_loop():
            local_notify_messages: List[BotForwardMessage] = []
            with self._notify_mutex:
                local_notify_messages = self._notify_messages
                self._notify_messages = []

            if len(local_notify_messages) == 0:
                return

            connected_channels: list = []
            connected_guilds = await self.__bot.rest.fetch_my_guilds()
            for guild in connected_guilds:
                channels = await self.__bot.rest.fetch_guild_channels(guild)
                for channel in channels:
                    if channel.type == hikari.ChannelType.GUILD_TEXT:
                        connected_channels.append(channel)

            for notify_message in local_notify_messages:
                embed = hikari.Embed(
                    title=notify_message.title,
                    description=f"{self._emoji_attention} {notify_message.message}",
                    color=hikari.colors.Color(self.__color_orange),
                )

                for channel in connected_channels:
                    try:
                        await self.__bot.rest.create_message(
                            channel=channel.id, embed=embed
                        )
                    except hikari.errors.ForbiddenError as exception:
                        logging.exception(exception)

    def notify(self, title: str, message: str) -> None:
        with self._notify_mutex:
            notify_message: BotForwardMessage = BotForwardMessage()
            notify_message.title = title
            notify_message.message = message
            self._notify_messages.append(notify_message)

    def start(self) -> None:
        try:
            self.__bot.run()
        except Exception as exception:
            logging.exception(exception)



#!/usr/bin/env python

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from multiprocessing import Lock
from typing import Dict, List

from nidibot.server_provider.game_server import GameServer


@dataclass
class BotForwardMessage:
    title: str = ""
    message: str = ""


@dataclass
class BackupDescription:
    readable_name: str = ""
    filepath: str = ""

    def __lt__(self, other):
        return self.readable_name < other.readable_name


@dataclass
class BotConfiguration:
    type: str = ""
    token: str = ""
    privileged_users: list = field(default_factory=list)
    allowed_channels: list = field(default_factory=list)
    notify_polling_seconds: int = 5


class BotBase(ABC):
    def __init__(self, configuration: BotConfiguration, game_servers: List[GameServer]):
        self._configuration = configuration
        self._game_servers = game_servers

        self._backups: Dict[str, List[BackupDescription]] = {}
        self._game_server_names: list = []
        for game_server in self._game_servers:
            self._game_server_names.append(game_server.name())
            self._backups[game_server.name()] = game_server.list_backups()

        self._notify_mutex = Lock()
        self._notify_messages: List[BotForwardMessage] = []

        self._emoji_no_access = "\U0001F925"
        self._emoji_ok = "\U00002705"
        self._emoji_attention = "\U000026A0"
        self._emoji_bad = "\U000026D4"
        self._emoji_unknown = "\U00002049\U0000FE0F"

    def _get_response_title(self, game_server: GameServer) -> str:
        server_status = game_server.status()

        title = f"{server_status.game_name}"
        if server_status.version:
            title += f" ({server_status.version})"

        title += f" - {server_status.address}"

        return title

    def _get_game_server(self, server_name: str = "") -> GameServer:
        if not server_name:
            game_server = self._game_servers[0]
        else:
            game_server = next(x for x in self._game_servers if x.name() == server_name)

        return game_server

    @abstractmethod
    def notify(self, title: str, message: str) -> None:
        pass

    @abstractmethod
    def start(self) -> None:
        pass



#!/usr/bin/env python

import asyncio
import logging
from datetime import date, datetime
from itertools import chain
from typing import List, Sequence

import nest_asyncio  # type: ignore
from telegram import BotCommand, ReplyKeyboardMarkup, ReplyKeyboardRemove, Update
from telegram.constants import ParseMode
from telegram.ext import (
    Application,
    CommandHandler,
    ContextTypes,
    ConversationHandler,
    MessageHandler,
)
from telegram.helpers import escape_markdown

from nidibot.bots.bot_base import BotBase, BotConfiguration, BotForwardMessage
from nidibot.server_provider.game_server import GameServer


class TelegramBot(BotBase):
    def __init__(self, configuration: BotConfiguration, game_servers: List[GameServer]):
        super().__init__(configuration=configuration, game_servers=game_servers)

        logging.getLogger("httpx").setLevel(logging.WARNING)
        logging.getLogger("httpcore").setLevel(logging.WARNING)
        logging.getLogger("telegram").setLevel(logging.WARNING)

        self.__bot = (
            Application.builder()
            .token(self._configuration.token)
            .post_init(self.__publish_commands)
            .build()
        )

        (
            self.__PROCESS_OPERATION,
            self.__BACKUP_RESTORE_SERVER,
            self.__BACKUP_RESTORE_FILEPATH,
            self.__CONVERSATION_END,
        ) = range(4)

        common_command_handler = ConversationHandler(
            allow_reentry=True,
            entry_points=[
                CommandHandler(
                    [
                        "status",
                        "start",
                        "stop",
                        "restart",
                        "backup_create",
                        "backup_list",
                    ],
                    self.__select_server,
                )
            ],
            states={
                self.__PROCESS_OPERATION: [
                    MessageHandler(
                        filters=None,
                        callback=self.__process_operation,
                    )
                ],
                self.__CONVERSATION_END: [
                    MessageHandler(filters=None, callback=self.__conversation_end)
                ],
            },
            fallbacks=[MessageHandler(filters=None, callback=self.__conversation_end)],
        )
        self.__bot.add_handler(common_command_handler)

        backup_restore_handler = ConversationHandler(
            allow_reentry=True,
            entry_points=[CommandHandler("backup_restore", self.__backup_restore)],
            states={
                self.__BACKUP_RESTORE_SERVER: [
                    MessageHandler(
                        filters=None,
                        callback=self.__backup_restore_server,
                    )
                ],
                self.__BACKUP_RESTORE_FILEPATH: [
                    MessageHandler(
                        filters=None,
                        callback=self.__backup_restore_filepath,
                    )
                ],
                self.__CONVERSATION_END: [
                    MessageHandler(filters=None, callback=self.__conversation_end)
                ],
            },
            fallbacks=[MessageHandler(filters=None, callback=self.__conversation_end)],
        )
        self.__bot.add_handler(backup_restore_handler)

        job_queue = self.__bot.job_queue
        job_queue.run_repeating(  # type: ignore
            self.__notify_loop,
            interval=self._configuration.notify_polling_seconds,
            first=0,
        )

    def __concatenate_sequences(
        self, sequence1: Sequence, sequence2: Sequence
    ) -> Sequence:
        sequence_type = type(sequence1)
        return sequence_type(chain(sequence1, sequence2))  # type: ignore

    async def __publish_commands(self, application: Application) -> None:
        commands = [
            BotCommand(
                "status", "Provides extended information about game server status."
            ),
            BotCommand(
                "start",
                "Starts server if it is offline, restarts server if it is online.",
            ),
            BotCommand("stop", "Stops server if it is online."),
            BotCommand(
                "restart",
                "Restarts server if it is online, starts server if it is offline.",
            ),
            BotCommand(
                "backup_create",
                "Creates backup of games server files and uploads them to storage.",
            ),
            BotCommand(
                "backup_list",
                "Lists available backups of specific game server.",
            ),
            BotCommand(
                "backup_restore",
                "Restores specific backup on a game server.",
            ),
        ]
        await application.bot.set_my_commands(commands)

    async def __select_server(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> int:
        if update.effective_user is None or update.effective_user.username is None:
            logging.critical("No username in incoming message!")
            return self.__CONVERSATION_END

        username = update.effective_user.username

        if update.message is None or update.message.text is None:
            logging.critical("No chat_id in incoming message!")
            return self.__CONVERSATION_END

        if (
            update.effective_message is None
            or update.effective_message.chat_id is None
            or update.effective_message.text is None
        ):
            logging.critical("No chat_id in incoming message!")
            return self.__CONVERSATION_END

        chat_id = update.effective_message.chat_id
        message_text = update.effective_message.text
        command = message_text.replace("/", "")[: message_text.find("@") - 1]

        if (
            len(self._configuration.allowed_channels) > 0
            and str(chat_id) not in self._configuration.allowed_channels
        ):
            logging.error(
                "Called '%s' by '%s' in not allowed channel '%s'.",
                command,
                username,
                chat_id,
            )
            return self.__CONVERSATION_END

        logging.debug("Called '%s' by '%s'.", command, username)

        if context.user_data is not None:
            context.user_data["command"] = command

        reply_keyboard = [[]]  # type: ignore
        for game_server in self._game_server_names:
            sub_keyboard = [[game_server]]
            reply_keyboard = self.__concatenate_sequences(reply_keyboard, sub_keyboard)  # type: ignore

        await update.message.reply_text(
            "Please select server:",
            reply_markup=ReplyKeyboardMarkup(
                reply_keyboard,
                one_time_keyboard=True,
                resize_keyboard=True,
            ),
        )

        return self.__PROCESS_OPERATION

    async def __process_operation(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> int:
        if update.message is None or update.message.from_user is None:
            logging.critical("No username in incoming message!")
            return self.__CONVERSATION_END

        username = update.message.from_user.username

        if update.message is None or update.message.text is None:
            logging.critical("No text in incoming message!")
            return self.__CONVERSATION_END

        server_name = update.message.text

        command = ""
        if context.user_data is not None:
            command = context.user_data["command"]

        if not command:
            return self.__CONVERSATION_END

        if update.effective_message is None or update.effective_message.chat_id is None:
            logging.critical("No chat_id in incoming message!")
            return self.__CONVERSATION_END

        chat_id = update.effective_message.chat_id

        logging.debug(
            "'%s' selected server '%s' for command '%s'.",
            username,
            server_name,
            command,
        )

        game_server = next(
            (x for x in self._game_servers if x.name() == server_name), None
        )
        if game_server is None:
            return self.__CONVERSATION_END

        #
        # Process un-privileged commands.
        #
        if command == "status":
            server_status = game_server.status()

            if server_status.status == "online":
                status_emoji = self._emoji_ok
            elif server_status.status == "offline":
                status_emoji = self._emoji_bad
            elif server_status.status == "restarting":
                status_emoji = self._emoji_attention
            else:
                status_emoji = self._emoji_unknown

            if server_status.update_available:
                update_emoji = self._emoji_attention
                update_text = "yes"
            else:
                update_emoji = self._emoji_ok
                update_text = "no"

            date_time = server_status.available_until.split(" ")
            date_parts = date_time[0].split("-")
            d0 = date(datetime.now().year, datetime.now().month, datetime.now().day)
            d1 = date(int(date_parts[0]), int(date_parts[1]), int(date_parts[2]))
            delta = d1 - d0
            days_left = f"({delta.days} days left)"

            server_name = self._get_response_title(game_server=game_server)
            response_text = f"__*{escape_markdown(server_name, version=2)}*__\n\n"
            response_text += (
                f"*Address:* {escape_markdown(server_status.address, version=2)}\n"
            )
            response_text += f"*Status:* {status_emoji} {server_status.status}\n"
            response_text += f"*Players:* {server_status.players_connected} / {server_status.players_limit}\n"
            response_text += f"*Available until:* {escape_markdown(server_status.available_until, version=2)} {escape_markdown(days_left, version=2)}\n"
            response_text += f"*Update available:* {update_emoji} {update_text}"

            await context.bot.send_message(
                chat_id,
                text=response_text,
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )

            return self.__CONVERSATION_END

        if command == "backup_list":
            self._backups[server_name] = game_server.list_backups()

            backup_sum_message = "Available backups:\n"
            for backup in self._backups[server_name]:
                backup_sum_message += (
                    f"\\- {escape_markdown(backup.readable_name, version=2)}\n"
                )

            await context.bot.send_message(
                chat_id,
                text=backup_sum_message,
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )
            return self.__CONVERSATION_END

        #
        # Process privileged commands.
        #
        if username not in self._configuration.privileged_users:
            await update.message.reply_text(
                f"Sorry but you don't have rights to call this command\\! {self._emoji_no_access}",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )
            return self.__CONVERSATION_END

        if command == "start":
            await context.bot.send_message(
                chat_id,
                text=f"{self._emoji_attention} Starting server\\!",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )

            game_server.start()

        elif command == "stop":
            await context.bot.send_message(
                chat_id,
                text=f"{self._emoji_attention} Stopping server\\!",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )

            game_server.stop()

        elif command == "restart":
            await context.bot.send_message(
                chat_id,
                text=f"{self._emoji_attention} Restarting server\\!",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )

            game_server.restart()

        elif command == "backup_create":
            await context.bot.send_message(
                chat_id,
                text=f"{self._emoji_attention} Started creating backup\\, please wait\\.",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )

            if game_server.create_backup():
                await context.bot.send_message(
                    chat_id,
                    text=f"{self._emoji_ok} Backup was created successfully\\!",
                    parse_mode=ParseMode.MARKDOWN_V2,
                    reply_markup=ReplyKeyboardRemove(),
                )
            else:
                await context.bot.send_message(
                    chat_id,
                    text=f"{self._emoji_bad} Backup creation failed\\, please check bot logs\\!",
                    parse_mode=ParseMode.MARKDOWN_V2,
                    reply_markup=ReplyKeyboardRemove(),
                )

        return self.__CONVERSATION_END

    async def __backup_restore(
        self, update: Update, _: ContextTypes.DEFAULT_TYPE
    ) -> int:
        if update.effective_user is None or update.effective_user.username is None:
            logging.critical("No username in incoming message!")
            return self.__CONVERSATION_END

        username = update.effective_user.username

        if update.message is None or update.message.text is None:
            logging.critical("No chat_id in incoming message!")
            return self.__CONVERSATION_END

        if update.effective_message is None or update.effective_message.chat_id is None:
            logging.critical("No chat_id in incoming message!")
            return self.__CONVERSATION_END

        chat_id = update.effective_message.chat_id

        if (
            len(self._configuration.allowed_channels) > 0
            and str(chat_id) not in self._configuration.allowed_channels
        ):
            logging.error(
                "Called 'backup_restore' by '%s' in not allowed channel '%s'.",
                username,
                chat_id,
            )
            return self.__CONVERSATION_END

        logging.debug("Called 'backup_restore' by '%s'.", username)

        if username not in self._configuration.privileged_users:
            await update.message.reply_text(
                f"Sorry but you don't have rights to call this command\\! {self._emoji_no_access}",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )
            return self.__CONVERSATION_END

        reply_keyboard = [[]]  # type: ignore
        for game_server in self._game_server_names:
            sub_keyboard = [[game_server]]
            reply_keyboard = self.__concatenate_sequences(reply_keyboard, sub_keyboard)  # type: ignore

        await update.message.reply_text(
            "Please select server:",
            reply_markup=ReplyKeyboardMarkup(
                reply_keyboard,
                one_time_keyboard=True,
                resize_keyboard=True,
            ),
        )

        return self.__BACKUP_RESTORE_SERVER

    async def __backup_restore_server(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> int:
        if update.message is None or update.message.from_user is None:
            logging.critical("No username in incoming message!")
            return self.__CONVERSATION_END

        username = update.message.from_user.username

        if update.message is None or update.message.text is None:
            logging.critical("No text in incoming message!")
            return self.__CONVERSATION_END

        server_name = update.message.text

        if context.user_data is not None:
            context.user_data["game_server"] = server_name

        logging.debug("'%s' selected server '%s'.", username, server_name)

        game_server = next(
            (x for x in self._game_servers if x.name() == server_name), None
        )
        if game_server is None:
            return self.__CONVERSATION_END

        backups = game_server.list_backups()
        if len(backups) > 0:
            reply_keyboard = [[]]  # type: ignore
            for backup_description in backups:
                sub_keyboard = [[backup_description.readable_name]]
                reply_keyboard = self.__concatenate_sequences(reply_keyboard, sub_keyboard)  # type: ignore

            await update.message.reply_text(
                "Please select backup:",
                reply_markup=ReplyKeyboardMarkup(
                    reply_keyboard,
                    one_time_keyboard=True,
                    resize_keyboard=True,
                ),
            )

            return self.__BACKUP_RESTORE_FILEPATH

        logging.warning("No backups available!")
        await update.message.reply_text(
            escape_markdown(
                text=f"{self._emoji_bad} No backups available!",
                version=2,
            ),
            reply_markup=ReplyKeyboardRemove(),
        )

        return self.__CONVERSATION_END

    async def __backup_restore_filepath(
        self, update: Update, context: ContextTypes.DEFAULT_TYPE
    ) -> int:
        if update.message is None or update.message.from_user is None:
            logging.critical("No username in incoming message!")
            return self.__CONVERSATION_END

        username = update.message.from_user.username

        if update.message is None or update.message.text is None:
            logging.critical("No text in incoming message!")
            return self.__CONVERSATION_END

        backup_readable_name = update.message.text

        logging.debug("'%s' selected backup '%s'.", username, backup_readable_name)

        if context.user_data is None:
            return self.__CONVERSATION_END

        server_name = context.user_data["game_server"]
        game_server = next(
            (x for x in self._game_servers if x.name() == server_name),
            None,
        )
        if game_server is None:
            return self.__CONVERSATION_END

        backups = self._backups[server_name]
        if backups is None:
            return self.__CONVERSATION_END

        backup_description = next(
            (x for x in backups if x.readable_name == backup_readable_name), None
        )
        if backup_description is None:
            return self.__CONVERSATION_END

        escaped_backup_name = escape_markdown(
            backup_description.readable_name, version=2
        )
        await update.message.reply_text(
            text=f"{self._emoji_attention} Started restoring backup from {escaped_backup_name}\\, please wait\\.",
            parse_mode=ParseMode.MARKDOWN_V2,
            reply_markup=ReplyKeyboardRemove(),
        )

        if game_server.restore_backup(backup_description.filepath):
            await update.message.reply_text(
                text=f"{self._emoji_ok} Backup from {escaped_backup_name} was restored successfully\\!",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )
        else:
            await update.message.reply_text(
                text=f"{self._emoji_bad} Restoring backup from {escaped_backup_name} failed\\, please check bot logs\\!",
                parse_mode=ParseMode.MARKDOWN_V2,
                reply_markup=ReplyKeyboardRemove(),
            )

        return ConversationHandler.END

    async def __conversation_end(
        self, _1: Update, _2: ContextTypes.DEFAULT_TYPE
    ) -> int:
        return ConversationHandler.END

    async def __notify_loop(self, context: ContextTypes.DEFAULT_TYPE) -> None:
        local_notify_messages: List[BotForwardMessage] = []
        with self._notify_mutex:
            local_notify_messages = self._notify_messages
            self._notify_messages = []

        if len(local_notify_messages) == 0:
            return

        for notify_message in local_notify_messages:
            for channel in self._configuration.allowed_channels:
                await context.bot.send_message(
                    channel,
                    text=f"__*{escape_markdown(text=notify_message.title, version=2)}*__"
                    f"\n{self._emoji_attention} {escape_markdown(text=notify_message.message, version=2)}",
                    parse_mode=ParseMode.MARKDOWN_V2,
                    reply_markup=ReplyKeyboardRemove(),
                )

    def notify(self, title: str, message: str) -> None:
        with self._notify_mutex:
            notify_message: BotForwardMessage = BotForwardMessage()
            notify_message.title = title
            notify_message.message = message
            self._notify_messages.append(notify_message)

    #
    # A bit of tricky solution to run python-telegram-bot in separate thread.
    #
    async def __start_bot(self) -> None:
        self.__bot.run_polling(
            allowed_updates=Update.ALL_TYPES,
            close_loop=False,
            stop_signals=None,
            drop_pending_updates=True,
        )

    def start(self) -> None:
        nest_asyncio.apply()

        try:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(self.__start_bot())
            loop.close()

        except Exception as exception:
            logging.exception(exception)
